# This is a sage code for rounding computations from flag algebras for paper on maximizing
# the induced density of C_5 by J. Balogh, P. Hu, B. Lidicky and F. Pfender.

import numpy

roundingDenominator = 10^10
eigVZero = 0.0000001
densityDenominator = 5040   # 7!

def read_problem_file(filename):
    f = open(filename, 'r')

    line = f.readline()
    line = line.strip()
    num_constraints = int(line)

    line = f.readline()
    line = line.strip()
    num_blocks = int(line)

    line = f.readline()
    parts = line.split()
    block_sizes = []
    for part in parts:
        block_sizes.append(int(part))

    blocks = []
    for size in block_sizes:
        if size < 0:
            block = numpy.zeros([num_constraints + 1, 1, -size], float)
        else:
            block = numpy.zeros([num_constraints + 1, size, size], float)
        blocks.append(block)

    line = f.readline()
    parts = line.split()
    constants = numpy.zeros([num_constraints], float)
    for i in xrange(num_constraints):
        constants[i] = float(parts[i])

    while True:
        line = f.readline()
        if len(line) == 0:
            break
        parts = line.split()
        if len(parts) == 0:
            continue
        constraint = int(parts[0])
        block_num = int(parts[1])
        row = int(parts[2])
        column = int(parts[3])
        value = float(parts[4])

        if block_sizes[block_num - 1] < 0:
            blocks[block_num - 1][constraint, 0, column - 1] += value
        else:
            blocks[block_num - 1][constraint, row - 1, column - 1] += value
            if row != column:
                blocks[block_num - 1][constraint, column - 1, row - 1] += value


    problem = dict()
    problem['num_constraints'] = num_constraints
    problem['num_blocks'] = num_blocks
    problem['block_sizes'] = block_sizes
    problem['blocks'] = blocks
    problem['constants'] = constants
    f.close()
    return problem

def read_solution_file(filename, problem):
    f = open(filename, 'r')

    num_constraints = problem['num_constraints']
    num_blocks = problem['num_blocks']
    block_sizes = problem['block_sizes']

    Y = numpy.zeros([num_constraints], float)
    line = f.readline()
    parts = line.split()
    for i in xrange(num_constraints):
        Y[i] = float(parts[i])

    Primal = []
    Dual = []

    for block_num in xrange(num_blocks):
        if block_sizes[block_num] < 0:
            Primal.append(numpy.zeros([1, -block_sizes[block_num]], float))
            Dual.append(numpy.zeros([1, -block_sizes[block_num]], float))
        else:
            Primal.append(numpy.zeros([block_sizes[block_num], block_sizes[block_num]], float))
            Dual.append(numpy.zeros([block_sizes[block_num], block_sizes[block_num]], float))

    while True:
        line = f.readline()
        if len(line) == 0:
            break
        parts = line.split()
        if len(parts) == 0:
            continue
        constraint = int(parts[0])
        block_num = int(parts[1])
        row = int(parts[2])
        column = int(parts[3])
        value = float(parts[4])

        if constraint == 2:
            block = Primal[block_num - 1]
        else:
            block = Dual[block_num - 1]

        if block_sizes[block_num - 1] < 0:
            block[0, column - 1] += value
        else:
            block[row - 1, column - 1] += value
            if row != column:
                block[column - 1, row - 1] += value


    solution = dict()
    solution['Y'] = Y
    solution['Primal'] = Primal
    solution['Dual'] = Dual
    return solution

def rational_approximation(V, den = roundingDenominator, strategy = "round"):
    if strategy == "round":
        f = lambda x: round(real_part(x))
    elif strategy == "floor":
        f = lambda x: floor(real_part(x))
    else:
        f = lambda x: ceil(real_part(x))

    best_denominator = den

    m = V.nrows()
    n = V.ncols()

    W = matrix(QQ, m, n)
    for i in xrange(m):
        for j in xrange(n):
            W[i,j] = float(f(V[i, j] * best_denominator))
    approximation = W / best_denominator
    
    return approximation

# round A
def rounding_matrix(A, out_eig = False):
    n = A.nrows()
    B = matrix(QQ, n)
    evA, eigA = A.eigenmatrix_left()

    for i in range(n):
        v = matrix(eigA[i])
        v = rational_approximation(v, roundingDenominator)
        v = v[0]
        z = v

        B[i] = z
        if abs(evA[i,i]) < eigVZero:
            evA[i,i] = 0
    if out_eig:
        return B[:n]
    else:
        evA = rational_approximation(evA, roundingDenominator)
        C = B.transpose() * evA * B
        return C


# test what bound those matrices can give for the problem
def testMatrices(problem, roundedMatrices):
    num_constraints = problem['num_constraints']
    num_blocks = problem['num_blocks']
    block_sizes = problem['block_sizes']
    blocks = problem['blocks']
    constants = problem['constants']
    min = -1
    for i in xrange(num_constraints):
        s = 0
        for j in xrange(num_blocks):
            if block_sizes[j] > 0:
                A = matrix(RDF, blocks[j][i+1])
                A = rational_approximation(A, densityDenominator)
                s += sum(sum(A.elementwise_product(roundedMatrices[j])))
        c = Rational(constants[i]) - s
        if min == -1 or c < min:
            min = c
    return min



problem = read_problem_file("C5_diff_L.dat-s")
solnMatrix = []

solution = read_solution_file("C5_diff_L.result", problem)
for block in solution['Primal']:
    solnMatrix.append(matrix(RDF, block))

roundedMatrices = []
roundedEigM = []
for i in xrange(problem['num_blocks']):
    if problem['block_sizes'][i] > 0:
        roundedMatrices.append(rounding_matrix(solnMatrix[i]))
        print "Matrix", i+1, "is rounded"
    else:
        m = solnMatrix[i].nrows()
        n = solnMatrix[i].ncols()
        roundedMatrices.append(matrix.zero(m,n))
        print "Matrix", i+1, "are slackness variables"
    

r = testMatrices(problem, roundedMatrices)
print r
print float(r)
